@@ -1,18 +1,28 @@
-""" Node
-
+"""
 A node in its simplest would retrieve a task from the central server by
 an API call, run this task and finally return the results to the central
 server again.
 
-The node application is seperated in 4 threads:
-- main thread, waits for new tasks to be added to the queue and
-    run the tasks
-- listening thread, listens for incommin websocket messages. Which
-    are handled by NodeTaskNamespace.
-- speaking thread, waits for results from docker to return and posts
-    them at the central server
-- proxy server thread, provides an interface for master containers
-    to post tasks and retrieve results
+The node application is runs 4 (Python-)threads:
+
+*Main thread*
+    Waits for new tasks to be added to the queue and run the tasks
+*Listening thread*
+    Listens for incommin websocket messages. Which are handled by
+    NodeTaskNamespace.
+*Speaking thread*
+    Waits for results from docker to return and posts them at the central
+    server
+*Proxy server thread*
+    Algorithm containers are isolated from the internet because of obvious
+    security reasons. The local-proxy-server provides an interface for
+    *master* containers to post tasks and retrieve their results.
+
+The node connects to the server using a websocket connection. This connection
+is mainly used for sharing status updates. This avoids the need for polling to
+see if there are new tasks available.
+
+-------------------------------------------------------------------------------
 """
 import sys
 import os
@@ -118,28 +128,21 @@ class NodeTaskNamespace(ClientNamespace):
 
 # ------------------------------------------------------------------------------
 class Node(object):
-    """Node to handle incomming computation requests.
-
-    The main steps this application follows: 1) retrieve (new) tasks
-    from the central server, 2) kick-off docker algorithm containers
-    based on this task and 3) retrieve the docker results and post
-    them to the central server.
-
-    TODO: read allowed repositories from the config file
     """
+    Authenticates to the central server, setup encrpytion, a
+    websocket connection, retrieving task that were posted while
+    offline, preparing dataset for usage and finally setup a
+    local proxy server..
 
-    def __init__(self, ctx):
-        """ Initialize a new Node instance.
+    Parameters
+    ----------
+    ctx:
+        Application context object.
 
-            Authenticates to the central server, setup encrpytion, a
-            websocket connection, retrieving task that were posted while
-            offline, preparing dataset for usage and finally setup a
-            local proxy server.
+    """
+    def __init__(self, ctx: Union[NodeContext, DockerNodeContext]):
 
-            :param ctx: application context, see utils
-        """
         self.log = logging.getLogger(logger_name(__name__))
-
         self.ctx = ctx
 
         # Initialize the node. If it crashes, shut down the parts that started
@@ -150,7 +153,8 @@ class Node(object):
             self.cleanup()
             raise
 
-    def initialize(self):
+    def initialize(self) -> None:
+        """Initialization of the node"""
         # check if docker is running, otherwise exit with error
         check_docker_running()
 
@@ -236,7 +240,7 @@ class Node(object):
 
         self.log.info('Init complete')
 
-    def __proxy_server_worker(self):
+    def __proxy_server_worker(self) -> None:
         """ Proxy algorithm container communcation.
 
             A proxy for communication between algorithms and central
@@ -290,7 +294,7 @@ class Node(object):
                 self.log.error('Proxyserver could not be started or crashed!')
                 self.log.error(e)
 
-    def sync_task_queue_with_server(self):
+    def sync_task_queue_with_server(self) -> None:
         """ Get all unprocessed tasks from the server for this node."""
         assert self.server_io.cryptor, "Encrpytion has not been setup"
 
@@ -302,13 +306,15 @@ class Node(object):
 
         self.log.info(f"received {self.queue._qsize()} tasks")
 
-    def __start_task(self, taskresult):
-        """Start a task.
+    def __start_task(self, taskresult: dict) -> None:
+        """
+        Start the docker image and notify the server that the task has been
+        started.
 
-            Start the docker image and notify the server that the task
-            has been started.
-
-            :param taskresult: an empty taskresult
+        Parameters
+        ----------
+        taskresult : dict
+            an 'empty' taskresult
         """
         task = taskresult['task']
         self.log.info("Starting task {id} - {name}".format(**task))
@@ -363,7 +369,7 @@ class Node(object):
                 f"node/{node_id}", json={"ip": node_ip}, method="PATCH"
             )
 
-    def __listening_worker(self):
+    def __listening_worker(self) -> None:
         """ Listen for incoming (websocket) messages from the server.
 
             Runs in a separate thread. Received events are dispatched
@@ -383,7 +389,7 @@ class Node(object):
                 self.log.error('Listening thread had an exception')
                 self.log.debug(e)
 
-    def __speaking_worker(self):
+    def __speaking_worker(self) -> None:
         """ Sending messages to central server.
 
             Routine that is in a seperate thread sending results
@@ -448,15 +454,16 @@ class Node(object):
                     }
                 )
             except Exception as e:
-                self.log.error('Speaking thread had an exception')
-                self.log.debug(e)
+                self.log.exception('Speaking thread had an exception')
 
-    def authenticate(self):
-        """ Authenticate to the central server
 
-            Authenticate with the server using the api-key. If the
-            server rejects for any reason we keep trying.
+    def authenticate(self) -> None:
         """
+        Authenticate with the server using the api-key from the configuration
+        file. If the server rejects for any reason -other than a wrong API key-
+        serveral attempts are taken to retry.
+        """
+
         api_key = self.config.get("api_key")
 
         success = False
@@ -488,7 +495,7 @@ class Node(object):
             self.log.critical('Unable to authenticate. Exiting')
             exit(1)
 
-    def private_key_filename(self):
+    def private_key_filename(self) -> Path:
         """Get the path to the private key."""
 
         # FIXME: Code duplication: vantage6/cli/node.py uses a lot of the same
@@ -507,7 +514,7 @@ class Node(object):
 
         return fullpath
 
-    def setup_encryption(self):
+    def setup_encryption(self) -> None:
         """Setup encryption ... or don't."""
         encrypted_collaboration = self.server_io.is_encrypted_collaboration()
         encrypted_node = self.config['encryption']["enabled"]
@@ -566,11 +573,9 @@ class Node(object):
             self.__tasks_dir = ctx.data_dir
             self.__vpn_dir = ctx.vpn_dir
 
-    def setup_vpn_connection(
-        self,
-        isolated_network_mgr: NetworkManager,
-        ctx: Union[DockerNodeContext, NodeContext]
-    ) -> VPNManager:
+    def setup_vpn_connection(self, isolated_network_mgr: NetworkManager,
+                             ctx: Union[DockerNodeContext, NodeContext]) \
+                                -> VPNManager:
         """
         Setup container which has a VPN connection
 
@@ -695,7 +700,7 @@ class Node(object):
             f.write(ovpn_config)
         return True
 
-    def link_docker_services(self):
+    def link_docker_services(self) -> None:
         docker_services = self.ctx.config.get("docker_services")
         if not docker_services:
             return
@@ -705,13 +710,11 @@ class Node(object):
                 container_name=container_name, config_alias=alias
             )
 
-    def connect_to_socket(self):
-        """ Create long-lasting websocket connection with the server.
-
-            The connection is used to receive status updates, such as
-            new tasks.
+    def connect_to_socket(self) -> None:
+        """
+        Create long-lasting websocket connection with the server. The
+        connection is used to receive status updates, such as new tasks.
         """
-
         self.socketIO = SocketIO(request_timeout=60)
 
         self.socketIO.register_namespace(NodeTaskNamespace('/tasks'))
@@ -737,12 +740,16 @@ class Node(object):
         self.log.info(f'Connected to host={self.server_io.host} on port='
                       f'{self.server_io.port}')
 
-    def get_task_and_add_to_queue(self, task_id):
-        """Fetches (open) task with task_id from the server.
-
-            The `task_id` is delivered by the websocket-connection.
+    def get_task_and_add_to_queue(self, task_id: int) -> None:
         """
+        Fetches (open) task with task_id from the server. The `task_id` is
+        delivered by the websocket-connection.
 
+        Parameters
+        ----------
+        task_id : int
+            Task identifyer
+        """
         # fetch (open) result for the node with the task_id
         tasks = self.server_io.get_results(
             include_task=True,
@@ -755,7 +762,7 @@ class Node(object):
         for task in tasks:
             self.queue.put(task)
 
-    def run_forever(self):
+    def run_forever(self) -> None:
         """Forever check self.queue for incoming tasks (and execute them)."""
         kill_listener = ContainerKillListener()
         try:
@@ -791,7 +798,7 @@ class Node(object):
             self.cleanup()
             sys.exit()
 
-    def cleanup(self):
+    def cleanup(self) -> None:
         if hasattr(self, 'socketIO') and self.socketIO:
             self.socketIO.disconnect()
         if hasattr(self, 'vpn_manager') and self.vpn_manager:
